// BalBot.ino
// UNO @ IDE 1.8.9

#include <Wire.h>
#include "I2Cdev.h"
#include <PID_v1.h>
#include <LMotorController.h>
#include "MPU6050_6Axis_MotionApps20.h"

// MOTOR CONTROLLER
#define ENA 5
#define IN1 6
#define IN2 7
#define ENB 10
#define IN3 8
#define IN4 9
#define MIN_ABS_SPEED 20 // compensate motor friction

// MPU control
MPU6050 mpu;
bool dmpReady = false;  
byte mpuIntStatus, devStatus;   
unsigned int packetSize;        
unsigned int fifoCount;     
byte fifoBuffer[64]; 

Quaternion q;           // rotation and acceleration data 
VectorFloat gravity;    // gravity data of MPU
float ypr[3];           // yaw, pitch and roll data

//PID
double setpoint = 180;  // vertival position
double PIDin, PIDout;

double Kp = 400.0;    // 0 ... 1000 default: 400
double Kd =  30.0;    // 0 ... 100 default:   30
double Ki = 200.0;    // 0 ... 500 default:  200

PID pid(&PIDin, &PIDout, &setpoint, Kp, Ki, Kd, DIRECT);

double motorSpeedFactorLeft  = 2.0;     // default: 2.0
double motorSpeedFactorRight = 2.0;     // default: 2.0

LMotorController motorController(ENA, IN1, IN2, ENB, IN3, IN4, motorSpeedFactorLeft, motorSpeedFactorRight);

volatile bool mpuInterrupt = false;     
void dmpDataReady()
{  mpuInterrupt = true;
}

void setup()
{   Serial.begin(250000);
    Wire.begin();
    TWBR = 24; // 400kHz I2C clock (200kHz if CPU is 8MHz)
    mpu.initialize();
    devStatus = mpu.dmpInitialize();

    mpu.setXGyroOffset(220); mpu.setYGyroOffset(75); mpu.setZGyroOffset(-90);
    mpu.setZAccelOffset(1800); // factory default

    if (devStatus == 0)
    {   Serial.println(F("MPU starting"));
        mpu.setDMPEnabled(true);
        attachInterrupt(0, dmpDataReady, RISING);
        mpuIntStatus = mpu.getIntStatus();
        Serial.println(F("MPU ready"));
        dmpReady = true;
        packetSize = mpu.dmpGetFIFOPacketSize();
        pid.SetMode(AUTOMATIC);
        pid.SetSampleTime(10);
        pid.SetOutputLimits(-255, 255);  
    }
    else
    {   Serial.print("DMP failed ");
        Serial.print(devStatus); 
    }
}

void loop()
{   if (!dmpReady) return;
    
    while (!mpuInterrupt && fifoCount < packetSize)
    {  pid.Compute();
       motorController.move(PIDout, MIN_ABS_SPEED); 
    }
    
    mpuInterrupt = false;
    mpuIntStatus = mpu.getIntStatus();
    fifoCount = mpu.getFIFOCount();

    if ((mpuIntStatus & 0x10) || fifoCount == 1024)
    {   mpu.resetFIFO();
        Serial.println(F("FIFO overflow!"));
    }
    else if (mpuIntStatus & 0x02)
    {   while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();
        mpu.getFIFOBytes(fifoBuffer, packetSize);
        fifoCount -= packetSize;
        mpu.dmpGetQuaternion(&q, fifoBuffer);
        mpu.dmpGetGravity(&gravity, &q);
        mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
              
        Serial.println(ypr[1] * 180/M_PI);      // actual tilt
               
        PIDin = ypr[1] * 180/M_PI + 180;
   }
}
